# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir.lowering_registry import LoweringRegistry
import operator
import ctypes
from numba_cuda_mlir.numba_cuda.types import ffi as _numba_ffi_type
from numba_cuda_mlir import types
from numba_cuda_mlir.lowering_utilities.type_conversions import (
    to_mlir_type,
    to_numba_type,
)
from numba_cuda_mlir.lowering_utilities import (
    convert,
    i64_of,
    DeferredLowering,
    index_of,
    storage_itemsize_bytes,
    is_complex_type as _is_complex_type,
    get_llvm_struct_for_complex as _get_llvm_struct_for_complex,
    complex_to_llvm_struct as _complex_to_llvm_struct,
    llvm_struct_to_complex as _llvm_struct_to_complex,
)
from numba_cuda_mlir._mlir.dialects import llvm, arith, memref
from numba_cuda_mlir.logging import trace
from numba_cuda_mlir._mlir.extras import types as T

registry = LoweringRegistry()
lower_getattr = registry.lower_getattr

llvm_kDynamic = (
    -2147483648
)  # std::numeric_limits<int32_t>::min(), ugh we need this in the bindings...


@registry.lower(ctypes.pointer, types.Array)
def lower_pointer(builder, target, args, kwargs):
    trace()
    args = builder.load_vars(args)
    value = args[0]
    value = convert(value, llvm.PointerType.get())
    builder.store_var(target, value)


@registry.lower_cast(types.Integer, types.Integer)
def lower_integer_cast(builder, target, args, kwargs):
    trace()
    args = builder.load_vars(args)
    value, to_type = args
    to_type = to_mlir_type(to_type)
    value = convert(value, to_type)
    builder.store_var(target, value)


@registry.lower_cast(types.CPointer, types.Integer)
def lower_pointer_to_int_cast(builder, target, args, kwargs):
    """Cast pointer to integer (ptrtoint)."""
    trace()
    ptr, to_type = builder.load_vars(args)
    result = llvm.ptrtoint(res=T.i64(), arg=ptr)
    to_type = to_mlir_type(to_type)
    result = convert(result, to_type)
    builder.store_var(target, result)


@registry.lower_cast(types.Integer, types.CPointer)
def lower_int_to_pointer_cast(builder, target, args, kwargs):
    """Cast integer to pointer (inttoptr)."""
    trace()
    int_val, _ = builder.load_vars(args)
    int_val = convert(int_val, T.i64())
    result = llvm.inttoptr(res=llvm.PointerType.get(), arg=int_val)
    builder.store_var(target, result)


@registry.lower(ctypes.cast, types.Any, types.Any)
def lower_cast(builder, target, args, kwargs):
    trace()
    args = builder.load_vars(args)
    value, to_type = args
    to_type = to_mlir_type(to_type)
    value = convert(value, to_type)
    builder.store_var(target, value)


@registry.lower(ctypes.POINTER, types.Any)
def lower_pointer(builder, target, args, kwargs):
    trace()
    args = builder.load_vars(args)
    value = args[0]
    nb_type = to_numba_type(value)
    nb_type = types.CPointer(nb_type)
    builder.store_var(target, nb_type)


@registry.lower(operator.setitem, types.CPointer, types.Number, types.Any)
def lower_pointer_setitem(builder, target, args, kwargs):
    trace()
    ptr, idx, value = builder.load_vars(args)
    # Get the numba type of the pointer - LLVM pointers are opaque
    nb_type = builder.get_numba_type(args[0])
    ele_ty = nb_type.dtype
    if ele_ty == types.none:
        raise TypeError(
            f"Cannot set item on pointer of type {nb_type}, must cast to a typed pointer first"
        )
    idx = convert(idx, T.i64())
    value_ty = builder.get_mlir_type(ele_ty)
    storage_ty = builder.get_storage_type(ele_ty)
    value = convert(value, value_ty)
    if _is_complex_type(value_ty):
        gep_ty = _get_llvm_struct_for_complex(value_ty)
        value = _complex_to_llvm_struct(value)
    else:
        gep_ty = storage_ty
        value = builder.as_storage(ele_ty, value)
    elementptr = llvm.getelementptr(
        llvm.PointerType.get(), ptr, [idx], [llvm_kDynamic], gep_ty, None
    )
    llvm.store(value, elementptr)


@registry.lower(operator.getitem, types.CPointer, types.Number)
def lower_pointer_getitem(builder, target, args, kwargs):
    trace()
    ptr, idx = builder.load_vars(args)
    nb_type = builder.get_numba_type(args[0])
    ele_ty = nb_type.dtype
    if ele_ty == types.none:
        raise TypeError(
            f"Cannot get item on pointer of type {nb_type}, must cast to a typed pointer first"
        )
    idx = convert(idx, T.i64())
    value_ty = builder.get_mlir_type(ele_ty)
    storage_ty = builder.get_storage_type(ele_ty)
    gep_ty = _get_llvm_struct_for_complex(value_ty) if _is_complex_type(value_ty) else storage_ty
    elementptr = llvm.getelementptr(
        llvm.PointerType.get(), ptr, [idx], [llvm_kDynamic], gep_ty, None
    )
    # Load the value from the computed address
    if _is_complex_type(value_ty):
        value = llvm.load(gep_ty, elementptr)
        value = _llvm_struct_to_complex(value, value_ty)
    else:
        value = llvm.load(storage_ty, elementptr)
        value = builder.from_storage(nb_type.dtype, value)
    builder.incref(nb_type.dtype, value)
    builder.store_var(target, value)


@registry.lower(operator.add, types.CPointer, types.Number)
@registry.lower(operator.iadd, types.CPointer, types.Number)
def lower_pointer_add(builder, target, args, kwargs):
    trace()
    ptr, num = builder.load_vars(args)
    nb_type = builder.get_numba_type(args[0])
    ele_ty = nb_type.dtype
    if ele_ty == types.none:
        raise TypeError(
            f"Cannot add to pointer of type {nb_type}, must cast to a typed pointer first"
        )
    w = storage_itemsize_bytes(ele_ty)
    num = convert(num, T.i64())
    ptri = llvm.ptrtoint(res=T.i64(), arg=ptr)
    ptri += num * w
    ptr = llvm.inttoptr(ptr.type, ptri)
    builder.store_var(target, ptr)


@registry.lower(operator.sub, types.CPointer, types.Number)
@registry.lower(operator.isub, types.CPointer, types.Number)
def lower_pointer_sub(builder, target, args, kwargs):
    trace()
    ptr, num = builder.load_vars(args)
    nb_type = builder.get_numba_type(args[0])
    ele_ty = nb_type.dtype
    if ele_ty == types.none:
        raise TypeError(
            f"Cannot subtract from pointer of type {nb_type}, must cast to a typed pointer first"
        )
    w = storage_itemsize_bytes(ele_ty)
    num = convert(num, T.i64())
    ptri = llvm.ptrtoint(res=T.i64(), arg=ptr)
    ptri -= num * w
    ptr = llvm.inttoptr(ptr.type, ptri)
    builder.store_var(target, ptr)


@registry.lower(types.ptr, types.CPointer)
def lower_aggregate_type_ptr(builder, target, args, kwargs):
    ptr = builder.load_var(args[0])
    assert "llvm.ptr" in str(ptr.type), f"Expected LLVM pointer, got {ptr.type}"
    builder.store_var(target, ptr)


@registry.lower(types.ptr, types.AggregateType)
def lower_aggregate_type_ptr(builder, target, args, kwargs):
    var = builder.load_var(args[0])
    with builder.alloca_insertion_point():
        fe_type = to_numba_type(var.type)
        attrs = types.get_numba_cuda_mlir_attributes(fe_type)
        alignment = attrs.get("align", None)
        ptr = llvm.alloca(llvm.PointerType.get(), i64_of(1), var.type, alignment=alignment)
    builder.store_var(target, ptr)


class DeferredFFIFromBuffer(DeferredLowering):
    def __call__(self, builder, target, args, kwargs):
        from numba_cuda_mlir._mlir import ir
        from numba_cuda_mlir.lowering_utilities import get_type_width

        array_value = builder.load_var(args[0])
        mr_type = ir.MemRefType(array_value.type)

        if mr_type.memory_space is not None:
            generic_mr_type = ir.MemRefType.get(
                shape=mr_type.shape,
                element_type=mr_type.element_type,
                layout=mr_type.layout,
            )
            array_value = memref.memory_space_cast(dest=generic_mr_type, source=array_value)
            mr_type = ir.MemRefType(array_value.type)

        ptr_as_index = memref.extract_aligned_pointer_as_index(array_value)

        md = memref.extract_strided_metadata(array_value)
        offset = index_of(md[1])

        elem_type = mr_type.element_type
        elem_bytes = get_type_width(elem_type) // 8

        byte_offset = arith.muli(
            offset,
            arith.constant(T.index(), elem_bytes),
        )
        data_ptr_index = arith.addi(ptr_as_index, byte_offset)

        ptr_as_i64 = arith.index_cast(T.i64(), data_ptr_index)
        ptr = llvm.inttoptr(res=llvm.PointerType.get(), arg=ptr_as_i64)
        builder.store_var(target, ptr)


@lower_getattr(_numba_ffi_type, "from_buffer")
def lower_ffi_from_buffer_getattr(_, mlir_lower, target, ffi_var):
    mlir_lower.store_var(target, DeferredFFIFromBuffer())



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Autogenerated by numba_cuda_mlir/decorators.py
# Please do not edit

from typing import Callable
from numba_cuda_mlir.descriptor import MLIRDispatcher

def jit(
    func_or_sig: Callable | str | None = None,
    abi: str = "numba",  # ABI for the compiled function
    capi: (
        bool | str
    ) = False,  # Use the function's name as the symbol name in the generated cubin. False means use the mangled name, true means use the function's name, and a string means use the string as the symbol name.
    chip: str = "sm_100",  # GPU compute capability used for compilation
    debug: bool = False,  # Emit debug information
    device: bool = False,  # Device-only functions can only be called by a kernel or other device functions
    dump: bool = False,  # Dump Numba IR and MLIR to stderr before compilation
    dump_ast: bool = False,  # Dump Python AST before/after transformations
    dump_ast_after_all: bool = False,  # Dump Python AST after each transformation pass
    dump_cubin: bool = False,  # Dump cubin to stderr after compilation
    dump_llvmir: bool = False,  # Dump LLVM IR to stderr before libnvvm compilation
    dump_mlir: bool = False,  # Dump MLIR to stderr before compilation
    dump_numba_ir: bool = False,  # Dump Numba IR to stderr before compilation
    dump_ptx: bool = False,  # Dump PTX to stderr before compilation
    experimental_ast_transforms: bool = False,  # Enable experimental AST transforms (consteval, loop unrolling)
    extensions: (
        list | None
    ) = None,  # List of argument handler extensions with prepare_args method
    fastmath: bool = False,  # Use faster approximations for floating-point arithmetic
    features: str = None,  # Additional PTX features to enable (for example, '+ptx80')
    inline: str | bool | Callable = "always",  # Inline strategy
    launch_bounds: (
        tuple | int | None
    ) = None,  # Launch bounds (max_threads_per_block, min_blocks_per_sm) or just max_threads_per_block as int
    lineinfo: bool = False,  # Emit line number information for profiling
    link: list = [],  # List of libraries to link against
    lto: bool = False,  # Enable link time optimization
    max_registers: int = None,  # Maximum number of registers per thread
    opt_level: int = 3,  # Optimization level
    print_after_all: bool = False,  # Print MLIR to stderr after every optimization pass
    print_before_all: bool = False,  # Print MLIR to stderr before every optimization pass
    ptxas_options: str = None,  # Additional options to pass to ptxas
    shared_memory_carveout: (
        str | int | None
    ) = None,  # Shared memory carveout percentage (-1 to 100, or 'default'/'maxl1'/'maxshared')
    signature: (
        tuple | list | str
    ) = None,  # Function signature to compile for. Setting to 'infer' will infer the signature from the type annotations of the arguments
) -> MLIRDispatcher:
    """
    numba_cuda_mlir JIT decorator. Use this function to decorate kernels and device functions.

    """



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Constant if-statement folding pass
import ast

from numba_cuda_mlir.ast_transforms.pipeline import ASTTransformPass, TransformContext


class ConstantIfTransformer(ast.NodeTransformer):
    """AST transformer that folds if statements with constant conditions."""

    def __init__(self):
        self.modified = False

    def visit_If(self, node: ast.If) -> ast.AST:
        # First visit children (including nested ifs in body/orelse)
        node = self.generic_visit(node)

        # Check if the condition is a constant
        if isinstance(node.test, ast.Constant):
            self.modified = True
            if node.test.value:
                # Condition is truthy - keep only the if body
                return node.body
            else:
                # Condition is falsy - keep only the else body
                return node.orelse if node.orelse else []

        return node


def transform_constant_if(tree: ast.Module) -> tuple[ast.Module, bool]:
    """Fold if statements with constant conditions.

    Returns (transformed_tree, was_modified).
    """
    transformer = ConstantIfTransformer()
    new_tree = transformer.visit(tree)
    ast.fix_missing_locations(new_tree)
    return new_tree, transformer.modified


class ConstantIfPass(ASTTransformPass):
    """Pipeline pass that folds if statements with constant conditions."""

    @property
    def name(self) -> str:
        return "ConstantIfFolding"

    def transform(self, tree: ast.Module, context: TransformContext) -> tuple[ast.Module, bool]:
        return transform_constant_if(tree)



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""Model classes representing a tensor comprehension.

These classes model the language more at an AST level as evaluated. Reasoning
about it typically involves processing this form into config objects that
represent actual op definitions (i.e. YAML).
"""

from typing import Any, Callable, Dict, List, Optional, Sequence, Set, Tuple
from enum import Enum

from ..... import ir as _ir
from .affine import *
from .scalar_expr import *
from .types import *
from .yaml_helper import *

###############################################################################
# Tensor expression nodes.
###############################################################################


class TensorExpression:
    """An expression that can appear on the RHS of a comprehension."""

    def to_scalar_expression(self) -> ScalarExpression:
        raise NotImplementedError()

    def visit_tensor_exprs(self, callback: Callable[["TensorExpression"], None]):
        """Visits all tensor expression reachable by the expression."""
        callback(self)

    def collect_dim_uses(self, uses: Set["DimDef"]):
        """Collects all DimDefs reachable through this expression."""

        def visit_dim_def(dim_def: AffineExprDef):
            if isinstance(dim_def, DimDef):
                uses.add(dim_def)

        def visit_affine_exprs(expr: "TensorExpression"):
            if isinstance(expr, TensorUse):
                for ind in expr.indices:
                    ind.visit_affine_exprs(visit_dim_def)
            if isinstance(expr, TensorReduceFn):
                for ind in expr.reduce_fn.reduce_dims:
                    ind.visit_affine_exprs(visit_dim_def)

        self.visit_tensor_exprs(visit_affine_exprs)

    def collect_tensor_uses(self, uses: Set["TensorUse"]):
        """Collects all TensorUses reachable through this expression."""

        def visit_tensor_use(expr: "TensorExpression"):
            if isinstance(expr, TensorUse):
                uses.add(expr)

        self.visit_tensor_exprs(visit_tensor_use)

    def collect_indices(self, indices: Set["index"]):
        """Collects all index accesses reachable through this expression."""

        def visit_index(expr: "TensorExpression"):
            if isinstance(expr, index):
                indices.add(expr)

        self.visit_tensor_exprs(visit_index)

    def collect_scalar_uses(self, uses: Set["ScalarDef"]):
        """Collects all ScalarDefs reachable through this expression."""

        def visit_scalar_def(expr: "TensorExpression"):
            if isinstance(expr, ScalarDef):
                uses.add(expr)

        self.visit_tensor_exprs(visit_scalar_def)

    def __add__(self, rhs: "TensorExpression") -> "TensorExpression":
        return BinaryFn.add(self, rhs)

    def __mul__(self, rhs) -> "TensorExpression":
        return BinaryFn.mul(self, rhs)

    def __sub__(self, rhs) -> "TensorExpression":
        return BinaryFn.sub(self, rhs)

    def __truediv__(self, rhs) -> "TensorExpression":
        return BinaryFn.div(self, rhs)

    def __hash__(self):
        return hash(id(self))


class TensorUse(TensorExpression):
    """A used tensor represented by its (tensor_name, indices).

    Note that forming a comprehension via direct assignment is performed through
    __setitem__ on the TensorDef level. However, performing a reduction with
    compound ops (+=, *=, etc) is done by doing a:
      TensorDef.__getitem__
      TensorUse.__iadd__
      TensorDef.__setitem__
    """

    def __init__(self, operand_def: "OperandDef", indices: Sequence[AffineExprDef]):
        self.operand_def = operand_def
        self.indices = tuple(indices)

    def to_scalar_expression(self) -> ScalarExpression:
        return ScalarArg(self.tensor_name).expr()

    @property
    def tensor_name(self) -> str:
        name = self.operand_def.name
        assert name is not None, "TensorDef not registered with an op"
        return name

    def _compute_reduce_dims(self, rhs: TensorExpression) -> Set[DimDef]:
        # Computes the reduction dims for implicit reductions. Assumes that the rhs
        # is the expression being reduced and self is being reduced into. Any
        # indices referenced on the rhs and not in self are considered reduction
        # dims and will be ordered as encountered on the rhs.
        rhs_dims = set()
        lhs_dims = set()
        rhs.collect_dim_uses(rhs_dims)
        self.collect_dim_uses(lhs_dims)
        return rhs_dims - lhs_dims

    def __iadd__(self, rhs: TensorExpression) -> "TensorReduceFn":
        return ReduceFnUse(BinaryFn.add, None, *self._compute_reduce_dims(rhs))(rhs)

    def __repr__(self):
        return (
            f"{self.operand_def.name}" f"[{', '.join([repr(i) for i in self.indices])}]"
        )


class TensorFn(TensorExpression):
    """Application of a tensor function."""

    def __init__(
        self,
        kind: "FunctionKind",
        name: Optional[str],
        operand_def: Optional["OperandDef"],
        type_var: Optional[TypeVar],
        args: Sequence[TensorExpression],
    ):
        if bool(name) + bool(operand_def) != 1:
            raise ValueError("One of 'name', 'operand_def' must be specified")
        self.name = name
        self.kind = kind
        self.operand_def = operand_def
        self.type_var = type_var
        self.args = args

    def to_scalar_expression(self) -> ScalarExpression:
        if self.operand_def:
            assert self.operand_def.name, "TensorFn not registered with an op"
        attr_name = self.operand_def.name if self.operand_def else None
        args = [arg.to_scalar_expression() for arg in self.args]
        return ScalarFn(self.kind, self.name, attr_name, self.type_var, args).expr()

    def visit_tensor_exprs(self, callback: Callable[["TensorExpression"], None]):
        super().visit_tensor_exprs(callback)
        for arg in self.args:
            arg.visit_tensor_exprs(callback)

    def __repr__(self):
        name = self.operand_def.name if self.operand_def else self.name
        return (
            f"{self.kind.name}.{name}(type_var={self.type_var}, "
            f"args={', '.join(repr(a) for a in self.args)})"
        )


class TensorReduceFn(TensorExpression):
    """Application of a reduction function.

    This captures the lhs (initial value) separately from the rhs.
    """

    def __init__(self, reduce_use: "ReduceFnUse", args: Sequence[TensorExpression]):
        self.reduce_use = reduce_use
        self.lhs = None  # type: Optional[TensorUse]
        self.args = args

    def to_scalar_expression(self) -> ScalarExpression:
        if self.lhs is None:
            raise ValueError(
                f"Cannot scalarize a TensorReduceFn that has not been "
                f"bound to its lhs: {self}"
            )
        full_args = [self.lhs.to_scalar_expression()] + [
            arg.to_scalar_expression() for arg in self.args
        ]
        fn_name = None
        attr_name = None
        if self.reduce_use.binary_fn:
            fn_name = self.reduce_use.binary_fn.fn_name
        if self.reduce_use.binary_attr:
            attr_name = self.reduce_use.binary_attr.operand_def.name
        return ScalarFn(FunctionKind.BINARY, fn_name, attr_name, None, full_args).expr()

    def visit_tensor_exprs(self, callback: Callable[["TensorExpression"], None]):
        for arg in self.args:
            arg.visit_tensor_exprs(callback)

    def __repr__(self):
        return f"{repr(self.reduce_use)}({', '.join(repr(a) for a in self.args)})"


class const(TensorExpression):
    """Returns the given constant floating point or integer value."""

    def __init__(self, value: Any):
        with _ir.Context():
            if isinstance(value, float):
                self.value = str(_ir.FloatAttr.get_f64(float(value)))
            elif isinstance(value, int):
                self.value = str(
                    _ir.IntegerAttr.get(_ir.IntegerType.get_signless(64), int(value))
                )
            else:
                raise ValueError(f"const requires int or float but got {type(value)}")

    def to_scalar_expression(self) -> ScalarExpression:
        return ScalarConst(self.value).expr()

    def __repr__(self):
        return f"const({self.value})"


class index(TensorExpression):
    """Returns the iteration index for a given dimension name.

    Resolves the given dimension name to obtain its position in the iteration
    domain of the operation.
    """

    def __init__(self, dim: DimDef):
        self.dim_def = dim
        self.dim = -1

    def resolve_dimension_name(self, affine_state: AffineBuildState):
        self.dim = affine_state.get_dim(self.dim_def.dimname)

    def to_scalar_expression(self) -> ScalarExpression:
        assert self.dim != -1, "Dimension name not resolved"
        return ScalarIndex(self.dim).expr()

    def __repr__(self):
        return f"index({repr(self.dim)})"


###############################################################################
# Function types and function definitions.
###############################################################################


class FunctionKind(Enum):
    UNARY = 0
    BINARY = 1
    TERNARY = 2
    TYPE = 3


class UnaryFnType:
    """Unary function.

    A unary function takes one tensor expression and returns the
    function evaluation result.
    """

    def __init__(self, fn_name: str):
        self.fn_name = fn_name

    def __call__(self, arg: TensorExpression) -> "TensorFn":
        return TensorFn(FunctionKind.UNARY, self.fn_name, None, None, [arg])

    def __repr__(self):
        return f"{self.fn_name}"


class UnaryFn:
    """Unary function namespace."""

    exp = UnaryFnType("exp")
    log = UnaryFnType("log")
    abs = UnaryFnType("abs")
    ceil = UnaryFnType("ceil")
    floor = UnaryFnType("floor")
    negf = UnaryFnType("negf")
    reciprocal = UnaryFnType("reciprocal")
    round = UnaryFnType("round")
    sqrt = UnaryFnType("sqrt")
    rsqrt = UnaryFnType("rsqrt")
    square = UnaryFnType("square")
    tanh = UnaryFnType("tanh")
    erf = UnaryFnType("erf")


class BinaryFnType:
    """Binary function.

    A binary function takes two tensor expressions and returns the
    function evaluation result.
    """

    def __init__(self, fn_name: str):
        self.fn_name = fn_name

    def __call__(self, arg0: TensorExpression, arg1: TensorExpression) -> "TensorFn":
        return TensorFn(FunctionKind.BINARY, self.fn_name, None, None, [arg0, arg1])

    def __repr__(self):
        return f"{self.fn_name}"


class BinaryFn:
    """Binary function namespace.

    As the integer types are signless, signedness is implement by different
    functions that treat integers as signed or unsigned values.

    Examples:
    - max -> `arith.MaxSIOp`
    - max_unsigned -> `arith.MaxUIOp`
    """

    add = BinaryFnType("add")
    sub = BinaryFnType("sub")
    mul = BinaryFnType("mul")
    div = BinaryFnType("div")
    div_unsigned = BinaryFnType("div_unsigned")
    max_signed = BinaryFnType("max_signed")
    min_signed = BinaryFnType("min_signed")
    max_unsigned = BinaryFnType("max_unsigned")
    min_unsigned = BinaryFnType("min_unsigned")
    powf = BinaryFnType("powf")


class TernaryFnType:
    """Ternary function.

    A ternary function takes three tensor expressions and returns the
    function evaluation result.
    """

    def __init__(self, fn_name: str):
        self.fn_name = fn_name

    def __call__(
        self, arg0: TensorExpression, arg1: TensorExpression, arg2: TensorExpression
    ) -> "TensorFn":
        return TensorFn(
            FunctionKind.TERNARY, self.fn_name, None, None, [arg0, arg1, arg2]
        )

    def __repr__(self):
        return f"{self.fn_name}"


class TernaryFn:
    """Ternary function namespace."""

    select = TernaryFnType("select")


class TypeFnType:
    """Type conversion function.

    A type conversion function takes a target type and a tensor expression and
    returns the casted tensor expression.
    """

    def __init__(self, fn_name: str):
        self.fn_name = fn_name

    def __call__(self, type_var: TypeVar, arg: TensorExpression) -> "TensorFn":
        return TensorFn(FunctionKind.TYPE, self.fn_name, None, type_var, [arg])

    def __repr__(self):
        return f"{self.fn_name}"


class TypeFn:
    """Type conversion function namespace.

    As the integer types are signless, signedness is implement by different cast
    functions that treat integers as signed (`cast_signed`) or unsigned
    (`cast_unsigned`) values.

    Examples:
    - cast_signed(I32 -> I64) -> `arith.ExtSIOp`
    - cast_unsigned(I32 -> I64) -> `arith.ExtUIOp`
    """

    cast_signed = TypeFnType("cast_signed")
    cast_unsigned = TypeFnType("cast_unsigned")


class ReduceFnUse:
    """Reduction function use.

    A reduction use specifies the reduction function and dimensions.
    """

    def __init__(
        self,
        binary_fn: Optional[BinaryFnType],
        binary_attr: Optional["BinaryFnAttrDef"],
        *reduce_dims: DimDef,
    ):
        if bool(binary_fn) + bool(binary_attr) != 1:
            raise ValueError("One of 'binary_fn', 'binary_attr' must be specified")
        self.binary_fn = binary_fn
        self.binary_attr = binary_attr
        self.reduce_dims = reduce_dims

    def __call__(self, *args: TensorExpression) -> "TensorReduceFn":
        return TensorReduceFn(self, args)

    def __repr__(self):
        fn = self.binary_fn if self.binary_fn else self.binary_attr
        return f"reduce_{repr(fn)}({', '.join(repr(d) for d in self.reduce_dims)})"


class ReduceFnType:
    """Reduction function.

    A binary function that reduces its RHS into its LHS.
    """

    def __init__(self, binary_fn: BinaryFnType):
        if not isinstance(binary_fn, BinaryFnType):
            raise ValueError(f"Reduce expected a BinaryFnType but got {binary_fn}")
        self.binary_fn = binary_fn

    def __getitem__(self, reduce_dims: Tuple[DimDef]) -> ReduceFnUse:
        return ReduceFnUse(self.binary_fn, None, *reduce_dims)

    def __repr__(self):
        return f"reduce_{repr(self.binary_fn)}"


class ReduceFn:
    add = ReduceFnType(BinaryFn.add)
    mul = ReduceFnType(BinaryFn.mul)
    max_signed = ReduceFnType(BinaryFn.max_signed)
    min_signed = ReduceFnType(BinaryFn.min_signed)
    max_unsigned = ReduceFnType(BinaryFn.max_unsigned)
    min_unsigned = ReduceFnType(BinaryFn.min_unsigned)


###############################################################################
# Operand definitions.
###############################################################################


class OperandKind(Enum):
    INPUT_TENSOR = 0
    SCALAR = 1
    OUTPUT_TENSOR = 2
    INDEX_ATTR = 3
    UNARY_FN_ATTR = 4
    BINARY_FN_ATTR = 5
    TERNARY_FN_ATTR = 6
    TYPE_FN_ATTR = 7


class OperandDef:
    """Definition of an operand passed to an operation.

    Keep the meta information of Tensor, Scalar, and Attribute operands and
    provide the shared registration functionality.
    """

    def __init__(
        self,
        kind: OperandKind,
        type_var: Optional[TypeVar] = None,
        size_exprs: Optional[Sequence[AffineExprDef]] = None,
        index_dims: Optional[Sequence[DimDef]] = None,
        default_indices: Optional[Sequence[int]] = None,
        default_fn: Optional[str] = None,
    ):
        if type_var and not isinstance(type_var, TypeVar):
            raise ValueError(f"OperandDef requires a TypeVar but got {repr(type_var)}")
        self.owner = None  # type: Optional["LinalgOpDef"]
        self.type_var = type_var
        self.size_exprs = size_exprs
        self.index_dims = index_dims
        self.default_indices = default_indices
        self.default_fn = default_fn
        self.kind = kind
        self.name = None  # type: Optional[str]
        self.registered_index = -1  # type: int

    def attach(self, index: int, name: str, owner: "LinalgOpDef"):
        if self.owner:
            raise ValueError(f"OperandDef already registered with an op: {self}")
        self.registered_index = index
        self.name = name
        self.owner = owner

    def is_input(self) -> bool:
        return self.kind == OperandKind.SCALAR or self.kind == OperandKind.INPUT_TENSOR

    def is_tensor(self) -> bool:
        return (
            self.kind == OperandKind.INPUT_TENSOR
            or self.kind == OperandKind.OUTPUT_TENSOR
        )

    def is_attribute(self) -> bool:
        return (
            self.kind == OperandKind.INDEX_ATTR
            or self.kind == OperandKind.UNARY_FN_ATTR
            or self.kind == OperandKind.BINARY_FN_ATTR
            or self.kind == OperandKind.TERNARY_FN_ATTR
            or self.kind == OperandKind.TYPE_FN_ATTR
        )

    def __hash__(self):
        return hash(id(self))

    def __repr__(self):
        return (
            f"{self.name}:OperandDef(kind={self.kind.name}, "
            f"type={repr(self.type_var)}, size_exprs={self.size_exprs}, "
            f"index_dims={self.index_dims}, "
            f"default_indices={self.default_indices}, "
            f"default_fn={self.default_fn})"
        )


class TensorDef:
    """Tensor operand definition.

    Tensor operands are indexed using the associated indexing_map when forwarded
    to the body of the structured op. A unique name identifies the tensor operands
    and an index determines their position in the operation's parameter list. A
    tensor definition takes type, a shape, and an optional flag to mark output
    tensors. Additionally, a tuple of index dimensions may be used to map the
    tensor to the loop dimensions of the operation. This mapping is needed to
    compute the indexing map of shape-only tensors that have no uses.
    """

    def __init__(
        self,
        type_var: TypeVar,
        *shape: AffineExprDef,
        index_dims: Optional[Sequence[DimDef]] = None,
        output: bool = False,
    ):
        if index_dims and len(shape) != len(index_dims):
            raise ValueError(
                f"Expected the shape rank {len(shape)} to match the "
                f"number of index_dims {len(index_dims)}"
            )
        if index_dims and any(not isinstance(dim, DimDef) for dim in index_dims):
            raise ValueError(
                f"TensorDef requires index dims of type DimDef but " f"got {index_dims}"
            )
        kind = OperandKind.OUTPUT_TENSOR if output else OperandKind.INPUT_TENSOR
        self.operand_def = OperandDef(
            kind, type_var=type_var, size_exprs=shape, index_dims=index_dims
        )

    def __getitem__(self, dims: Sequence[AffineExprDef]) -> TensorUse:
        assert self.operand_def.owner, "TensorDef is not registered with an op"
        state = AffineBuildState(
            global_state=self.operand_def.owner._affine_state, allow_new_symbols=False
        )
        if not isinstance(dims, tuple):
            dims = (dims,)  # Handle single subscript case.
        # Special case: (None) is a 0d-scalar use.
        if dims == (None,):
            dims = ()

        exprs = []
        for expr_def in dims:
            if not isinstance(expr_def, AffineExprDef):
                raise KeyError(
                    "A TensorDef can only be subscripted by a tuple of affine dims"
                )
            exprs.append(expr_def)
        return TensorUse(self.operand_def, exprs)

    def __setitem__(self, dims: Sequence[AffineExprDef], value: TensorExpression):
        """Creates a new 1:1 comprehension by binding this tensor to an expression.

        Note that due to the way assignment works in Python, we have to capture
        direct assignment as a setitem on the TensorDef.
        """
        if not isinstance(value, TensorExpression):
            raise ValueError(
                f"Only TensorExpressions can be assigned to TensorDefs. "
                f"Got: {repr(value)}"
            )
        use = self[dims]
        comp = Comprehension((use, value))
        self.operand_def.owner.comprehensions.append(comp)


class ScalarDef(TensorExpression):
    """Scalar operand definition.

    Scalar operands are forwarded to the body of the structured op as they are.
    A unique name identifies the scalars and an index determines their position in
    the operation's parameter list.
    """

    def __init__(self, type_var: TypeVar):
        self.operand_def = OperandDef(OperandKind.SCALAR, type_var=type_var)

    @property
    def scalar_name(self) -> str:
        name = self.operand_def.name
        assert name is not None, "ScalarDef not registered with an op"
        return name

    def to_scalar_expression(self) -> ScalarExpression:
        return ScalarArg(self.scalar_name).expr()


class IndexAttrDef:
    """Index attribute definition.

    Index attributes provide a way to define and set symbols that can be used in
    indexing expressions. Every attribute specifies a tuple of symbols that at
    compile-time are replaced by integer values as well as their default values.
    """

    def __init__(self, *sizes: SymbolDef, default: Sequence[int]):
        if any(not isinstance(size, SymbolDef) for size in sizes):
            raise ValueError(
                f"IndexAttrDef requires sizes of type SymbolDef " f"but got {sizes}"
            )
        if any(not isinstance(default_val, int) for default_val in default):
            raise ValueError(
                f"IndexAttrDef requires default values of type int "
                f"but got {default}"
            )
        if len(sizes) != len(default):
            raise ValueError(
                f"IndexAttrDef expects {len(sizes)} default values "
                f"but got {len(default)}"
            )
        self.operand_def = OperandDef(
            OperandKind.INDEX_ATTR, size_exprs=sizes, default_indices=default
        )


class UnaryFnAttrDef:
    """Unary function attribute definition.

    Unary function attributes provide a way to make the arithmetic computation
    parametrizable. Every attribute specifies a default unary function
    that may be overwritten at operation instantiation time.
    """

    def __init__(self, default: "UnaryFnType"):
        if not isinstance(default, UnaryFnType):
            raise ValueError(
                f"UnaryFnAttrDef requires default of type UnaryFnType "
                f"but got {default}"
            )
        self.operand_def = OperandDef(
            OperandKind.UNARY_FN_ATTR, default_fn=default.fn_name
        )

    def __call__(self, arg: TensorExpression) -> TensorFn:
        return TensorFn(FunctionKind.UNARY, None, self.operand_def, None, [arg])


class BinaryFnAttrDef:
    """Binary function attribute definition.

    Binary function attributes provide a way to make the arithmetic computation
    parametrizable. Every attribute specifies a default binary function
    that may be overwritten at operation instantiation time.
    """

    def __init__(self, default: "BinaryFnType"):
        if not isinstance(default, BinaryFnType):
            raise ValueError(
                f"BinaryFnAttrDef requires default of type BinaryFnType "
                f"but got {default}"
            )
        self.operand_def = OperandDef(
            OperandKind.BINARY_FN_ATTR, default_fn=default.fn_name
        )

    def __call__(self, arg0: TensorExpression, arg1: TensorExpression) -> TensorFn:
        return TensorFn(FunctionKind.BINARY, None, self.operand_def, None, [arg0, arg1])

    def __getitem__(self, reduce_dims: Tuple[DimDef]) -> ReduceFnUse:
        return ReduceFnUse(None, self, *reduce_dims)


class TernaryFnAttrDef:
    """Ternary function attribute definition.

    Ternary function attributes provide a way to make the arithmetic computation
    parametrizable. Every attribute specifies a default Ternary function
    that may be overwritten at operation instantiation time.
    """

    def __init__(self, default: "TernaryFnType"):
        if not isinstance(default, TernaryFnType):
            raise ValueError(
                f"TernaryFnAttrDef requires default of type TernaryFnType "
                f"but got {default}"
            )
        self.operand_def = OperandDef(
            OperandKind.TERNARY_FN_ATTR, default_fn=default.fn_name
        )

    def __call__(self, arg0: TensorExpression, arg1: TensorExpression) -> TensorFn:
        return TensorFn(
            FunctionKind.TERNARY, None, self.operand_def, None, [arg0, arg1]
        )

    def __getitem__(self, reduce_dims: Tuple[DimDef]) -> ReduceFnUse:
        return ReduceFnUse(None, self, *reduce_dims)


class TypeFnAttrDef:
    """Type conversion function attribute definition.

    Type conversion function attributes provide a way to make type conversions
    parameterizable. Every attribute specifies a default type conversion function
    that may be overwritten at operation instantiation time.
    """

    def __init__(self, default: "TypeFnType"):
        if not isinstance(default, TypeFnType):
            raise ValueError(
                f"TypeFnAttrDef requires default of type TypeFnType "
                f"but got {default}"
            )
        self.operand_def = OperandDef(
            OperandKind.TYPE_FN_ATTR, default_fn=default.fn_name
        )

    def __call__(self, type_var: TypeVar, arg: TensorExpression) -> TensorFn:
        return TensorFn(FunctionKind.TYPE, None, self.operand_def, type_var, [arg])


###############################################################################
# Operation definition.
###############################################################################


class Comprehension:
    """Represents a single comprehension."""

    def __init__(self, *bindings: Tuple[TensorUse, TensorExpression]):
        self.definitions = list()  # List[TensorUse]
        self.values = list()  # List[TensorExpression]

        # Find the lhs to reduction rhs.
        for assign, value in bindings:
            if isinstance(value, TensorReduceFn):
                if value.lhs:
                    raise ValueError(f"Reduction expression already assigns: {value}")
                value.lhs = assign
            self.definitions.append(assign)
            self.values.append(value)

    @property
    def all_reduction_dims(self) -> Set[Tuple[DimDef, ...]]:
        """Gets the reduction dims for the comprehension or None."""
        result = set()
        for use in self.values:
            if isinstance(use, TensorReduceFn):
                result.add(use.reduce_use.reduce_dims)
            else:
                result.add(tuple())
        return result

    def __repr__(self):
        if len(self.definitions) > 1:
            defs_repr = f"({', '.join(repr(d) for d in self.definitions)})"
            values_repr = f"({', '.join(repr(v) for v in self.values)})"
        else:
            defs_repr = f"{repr(self.definitions[0])}"
            values_repr = f"{repr(self.values[0])}"

        return f"{defs_repr} = {values_repr}"


class OpInterfaceDef:
    """An interface that an op implements."""

    def __init__(self, cpp_name: str):
        self.cpp_name = cpp_name


ContractionOpInterface = OpInterfaceDef("LinalgContractionOpInterface")
ConvolutionOpInterface = OpInterfaceDef("LinalgConvolutionOpInterface")
FillOpInterface = OpInterfaceDef("LinalgFillOpInterface")


class OpDefinitionDef:
    """A method that an op implements."""

    def __init__(self, def_name: str):
        self.def_name = def_name


Canonicalizer = OpDefinitionDef("hasCanonicalizer")


class OpMetadataDef(YAMLObject):
    """Metadata about the op (generally not behavior impacting)."""

    yaml_tag = "!LinalgOpMetadata"

    def __init__(self, name: str, cpp_class_name: Optional[str], doc: Optional[str]):
        self.name = name
        self.cpp_class_name = cpp_class_name if cpp_class_name is not None else name
        self.doc = doc
        self.implements = []  # type: List[OpInterfaceDef]
        self.defines = []  # type: List[OpDefinitionsDef]

    def to_yaml_custom_dict(self):
        d = dict(
            name=self.name,
            cpp_class_name=self.cpp_class_name,
            doc=self.doc,
        )
        if self.implements:
            d["implements"] = [intr.cpp_name for intr in self.implements]
        if self.defines:
            d["defines"] = [defi.def_name for defi in self.defines]
        return d


class LinalgOpDef:
    """Definition of a linalg op."""

    def __init__(
        self, name: str, cpp_class_name: Optional[str] = None, doc: Optional[str] = None
    ):
        self.metadata = OpMetadataDef(name=name, cpp_class_name=cpp_class_name, doc=doc)
        self.registered_operands = dict()  # type: Dict[str, OperandDef]
        self.domain = list()  # type: List[DimDef]
        self.comprehensions = list()  # type: List[Comprehension]
        self._affine_state = AffineBuildState()

    def add_operand(self, name: str, operand: OperandDef):
        """Registers an operand."""
        if name in self.registered_operands:
            raise ValueError(
                f"The operand {name} is already registered "
                f"to {self.registered_operands['name']}"
            )
        structured_op_methods = [
            "inputs",
            "outputs",
            "result_tensors",
            "region",
            "iterator_types",
            "indexing_maps",
            "getRegionBuilder",
            "getLibraryCallName",
        ]
        if operand.is_attribute() and name in structured_op_methods:
            raise ValueError(
                f"The attribute name {name} conflicts with a structured "
                f"op method name"
            )
        # Ensure output tensors are registered after input tensors and scalars and
        # attributes are registered after all other operand types.
        if operand.is_input() and any(
            not op_def.is_input() for op_def in self.registered_operands.values()
        ):
            raise ValueError(f"Input {name} registered after an output or attribute")
        if operand.kind == OperandKind.OUTPUT_TENSOR and any(
            op_def.is_attribute() for op_def in self.registered_operands.values()
        ):
            raise ValueError(f"Output {name} registered after an attribute")
        operand.attach(len(self.registered_operands), name, self)
        self.registered_operands[name] = operand

    def __repr__(self):
        lines = [f"LinalgOpDef({self.metadata.name} -> {self.metadata.cpp_class_name},"]
        for name, operand in self.registered_operands.items():
            lines.append(f"  {operand}")
        if self.comprehensions:
            lines[-1] += " {"
            for comprehension in self.comprehensions:
                lines.append(f"    {comprehension}")
            lines.append("}")
        return "\n".join(lines)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import inspect
import pprint
import os
from typing import Any, Dict, Callable

import warnings
from numba_cuda_mlir.numba_cuda import compiler, types, typing
from numba_cuda_mlir.numba_cuda.core import errors, funcdesc, sigutils
from numba_cuda_mlir.numba_cuda.core.compiler_lock import global_compiler_lock
from numba_cuda_mlir.numba_cuda.compiler import CompilerBase, DefaultPassBuilder
from numba_cuda_mlir.numba_cuda.flags import CUDAFlags
from numba_cuda_mlir.numba_cuda.core.options import ParallelOptions
from numba_cuda_mlir.numba_cuda.core.errors import (
    NumbaDebugInfoWarning,
)
from numba_cuda_mlir.numba_cuda.compiler import compile_result
from numba_cuda_mlir.descriptor import mlir_target
from numba_cuda_mlir.lowering_utilities.type_conversions import to_numba_type
from numba_cuda_mlir.numba_cuda.core.compiler_machinery import (
    FunctionPass,
    LoweringPass,
    PassManager,
    register_pass,
)
from numba_cuda_mlir.numba_cuda.core.typed_passes import (
    AnnotateTypes,
    IRLegalization,
    NopythonTypeInference,
)
from numba_cuda_mlir.numba_cuda.core.typed_passes import (
    fallback_context,
    type_inference_stage,
)

# numba-cuda forked numba.core.ir into numba.cuda.core.ir, creating a separate
# UndefinedType class and UNDEFINED singleton. The core SSA pass inserts
# numba.core.ir.UNDEFINED into phi nodes, but numba-cuda's typeinfer checks
# `iv is not ir.UNDEFINED` using its own copy — the identity check fails.
# Unify them so phi nodes with undefined incoming values are handled correctly.
import numba_cuda_mlir.numba_cuda.core.ir as _cuda_ir

from numba_cuda_mlir.numba_cuda.typing.typeof import Purpose, typeof
from numba_cuda_mlir.numba_cuda.core.untyped_passes import (
    LiteralUnroll,
    InlineInlinables,
)
from numba_cuda_mlir.numbair_transforms import (
    NumbaCudaMlirLiteralUnroll,
    NumbaCudaMlirInlineInlinables,
    PostInlineWholeFunctionPlanners,
)

import numba_cuda_mlir.mlir_lowering as lowering
import numba_cuda_mlir.mlir_optimization as opt
from numba_cuda_mlir._launch_config import (
    _LAUNCH_CONFIG_TRACKER_METADATA_KEY,
    _LAUNCH_CONFIG_TRACKER_OPTION,
)
from numba_cuda_mlir.decorators import mlir_jit
from numba_cuda_mlir.ast_transforms import apply_ast_transforms
from numba_cuda_mlir.errors import (
    InternalCompilerError,
    UserFacingInternalCompilerError,
    handle_lowering_error,
)


@register_pass(mutates_CFG=True, analysis_only=False)
class MLIRTypeInference(FunctionPass):
    _name = "mlir_type_inference"
    _raise_errors = True

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        """
        Type inference and legalization
        """
        with fallback_context(
            state, 'Function "%s" failed type inference' % (state.func_id.func_name,)
        ):
            # Type inference
            typemap, return_type, calltypes, errs = type_inference_stage(
                state.typingctx,
                state.targetctx,
                state.func_ir,
                state.args,
                state.return_type,
                state.locals,
                raise_errors=self._raise_errors,
            )
            state.typemap = typemap
            # save errors in case of partial typing
            state.typing_errors = errs
            if self._raise_errors:
                state.return_type = return_type
            state.calltypes = calltypes

        def legalize_return_type(return_type, interp, targetctx):
            """
            Reject function object return types.
            """

            if isinstance(return_type, types.Function) or isinstance(return_type, types.Phantom):
                if self._raise_errors:
                    msg = "Can't return function object ({}) in nopython mode"
                    raise errors.NumbaTypeError(msg.format(return_type))

        with fallback_context(
            state, 'Function "%s" has invalid return type' % (state.func_id.func_name,)
        ):
            legalize_return_type(state.return_type, state.func_ir, state.targetctx)
        return True


class MLIRLibrary:
    def __init__(self, code, ptx="", mlir_str=""):
        self._code = code
        self._ptx = ptx
        self._mlir_str = mlir_str
        # Objects that need to be kept alive for the lifetime of the kernels
        # generated by this library, e.g., device arrays captured from global
        # scope whose pointers are baked into the generated PTX as constants.
        self.referenced_objects = {}

    def get_ir(self):
        return self._code

    def get_llvm_str(self):
        return self._mlir_str

    def get_asm_str(self, cc=None):
        return self._ptx


# "mutates_cfg" probably carries more semantic load than just whether the CFG
# has been mutated, so we set it to True here to ensure Numba doesn't hold some
# invalid assumptions
@register_pass(mutates_CFG=True, analysis_only=False)
class MLIRBackend(LoweringPass):
    _name = "mlir_backend"

    def __init__(self):
        LoweringPass.__init__(self)

    def run_pass(self, state):
        """
        Back-end: Packages lowering output in a compile result
        """
        signature = typing.signature(state.return_type, *state.args)
        targetctx = state.targetctx
        func_ir = state.func_ir
        typemap = state.typemap
        restype = state.return_type
        calltypes = state.calltypes
        flags = state.flags
        metadata = state.metadata
        fndesc = funcdesc.PythonFunctionDescriptor.from_specialized_function(
            func_ir,
            typemap,
            restype,
            calltypes,
            mangler=targetctx.mangler,
            inline=flags.forceinline,
            noalias=flags.noalias,
            abi_tags=[flags.get_mangle_string()],
        )

        code = (
            f"=============== Numba IR ===============\n\n"
            f"{func_ir.dump_to_string()}\n\n"
            f"=============== Numba IR Type Map ===============\n\n"
            f"{pprint.pformat(typemap)}\n\n"
        )

        target_options = metadata["targetoptions"]
        dump = target_options.get("dump_numba_ir", False) or target_options.get("dump", False)

        if dump:
            print(code)

        library = MLIRLibrary(code)

        lower = lowering.MLIRLower(targetctx, fndesc, func_ir, metadata=metadata)

        try:
            with targetctx.push_code_library(library):
                lower.lower_to_mlir()
                lower.verify_mlir_module()
        except UserFacingInternalCompilerError:
            raise
        except (NotImplementedError, InternalCompilerError) as exc:
            if os.environ.get("NUMBA_CUDA_MLIR_ICE_FULL_TB", "0").strip() == "1":
                raise exc
            handle_lowering_error(lower, func_ir)

        state.cr = compile_result(
            typing_context=state.typingctx,
            target_context=state.targetctx,
            typing_error=state.status.fail_reason,
            type_annotation=state.type_annotation,
            library=library,
            call_helper=None,
            signature=signature,
            fndesc=fndesc,
            metadata=metadata,
        )
        return True


def get_compiler_class(
    targetoptions: Dict[str, Any],
    launch_config_tracker=None,
):
    class MLIRCompiler(CompilerBase):
        def define_pipelines(self):
            dpb = DefaultPassBuilder
            pm = PassManager("mlir")

            untyped_passes = dpb.define_untyped_pipeline(self.state)
            # Replace passes with numba_cuda_mlir-specific versions:
            # - LiteralUnroll -> NumbaCudaMlirLiteralUnroll (supports both literal_unroll versions)
            # - InlineInlinables -> NumbaCudaMlirInlineInlinables (skips self-recursive functions)
            modified_passes = []
            for impl, desc in untyped_passes.passes:
                if impl is LiteralUnroll:
                    modified_passes.append((NumbaCudaMlirLiteralUnroll, desc))
                elif impl is InlineInlinables:
                    modified_passes.append((NumbaCudaMlirInlineInlinables, desc))
                    modified_passes.append(
                        (
                            PostInlineWholeFunctionPlanners,
                            "post-inline whole-function extension planners",
                        )
                    )
                else:
                    modified_passes.append((impl, desc))
            pm.passes.extend(modified_passes)

            typed_passes = dpb.define_typed_pipeline(self.state)

            def replace_type_inference_pass(implementation, description):
                if implementation is NopythonTypeInference:
                    return (MLIRTypeInference, "mlir frontend")
                else:
                    return (implementation, description)

            mlir_typed_passes = [
                replace_type_inference_pass(implementation, description)
                for implementation, description in typed_passes.passes
            ]

            pm.passes.extend(mlir_typed_passes)

            lowering_passes = self.define_mlir_lowering_pipeline(self.state)
            pm.passes.extend(lowering_passes.passes)

            pm.finalize()
            return [pm]

        def define_mlir_lowering_pipeline(self, state):
            pm = PassManager("mlir_lowering")
            # Numba legalization
            pm.add_pass(IRLegalization, "ensure IR is legal prior to lowering")
            pm.add_pass(AnnotateTypes, "annotate types")

            # Lowering to MLIR
            pm.add_pass(MLIRBackend, "Run MLIR backend")

            pm.finalize()
            return pm

        # Capture user targetoptions from the active stack into compiler metadata
        def __init__(
            self,
            typingctx,
            targetctx,
            library,
            args,
            return_type,
            flags,
            locals,
            call_conv=None,
            abi_info=None,
        ):
            # These are added to work with the vendored Numba/Numba-CUDA code.
            # They will eventually need implementing, as part of the
            # requirement to support them in line with Numba-CUDA, but for now
            # we accept these parameters if they are None so that callees can
            # correctly call it.
            if call_conv is not None:
                raise RuntimeError("MLIR compiler does not support call_conv")
            if abi_info is not None:
                raise RuntimeError("MLIRCompiler does not support abi_info")
            super().__init__(typingctx, targetctx, library, args, return_type, flags, locals)
            # Attach options early so all passes can see them via state.metadata
            self.state.metadata["targetoptions"] = targetoptions
            if launch_config_tracker is not None:
                self.state.metadata[_LAUNCH_CONFIG_TRACKER_METADATA_KEY] = launch_config_tracker

    return MLIRCompiler


@global_compiler_lock
def compile_mlir(pyfunc, return_type, args, targetoptions: Dict[str, Any]):
    from numba_cuda_mlir.install_registry import register_lowering
    from numba_cuda_mlir.tools import resolve_gpu_target

    register_lowering()
    launch_config_tracker = targetoptions.pop(_LAUNCH_CONFIG_TRACKER_OPTION, None)
    gpu_target = resolve_gpu_target(targetoptions)
    targetoptions["chip"] = gpu_target["chip"]

    # Apply AST transforms at compile time with argtypes available
    # This allows consteval to access argument types and target options
    pyfunc, transformed_source = apply_ast_transforms(pyfunc, targetoptions, args)

    typingctx = mlir_target.typing_context
    targetctx = mlir_target.target_context

    # Do not compile to native code
    flags = CUDAFlags()
    flags.no_compile = True
    flags.no_cpython_wrapper = True
    flags.no_cfunc_wrapper = True

    from numba_cuda_mlir.memory_management.config import is_nrt_enabled

    flags.nrt = is_nrt_enabled()

    flags.auto_parallel = ParallelOptions(targetoptions.get("parallel", False))
    flags.nvvm_options = {}
    is_lto = targetoptions.get("lto", False) or targetoptions.get("_compile_output") == "ltoir"
    if is_lto:
        flags.nvvm_options["gen-lto"] = None

    # Both debug and lineinfo turn on debug information in the compiled code,
    # but we keep them separate arguments in case we later want to overload
    # some other behavior on the debug flag. In particular, -opt=3 is not
    # supported with debug enabled, and enabling only lineinfo should not
    # affect the error model.
    flags.dbg_directives_only = targetoptions.get("lineinfo", False)
    flags.debuginfo = targetoptions.get("debug", False) or flags.dbg_directives_only

    if flags.debuginfo:
        try:
            inspect.getsource(pyfunc)
        except OSError:
            warnings.warn(
                NumbaDebugInfoWarning(
                    f"Could not find source for function: {pyfunc}. "
                    "Debug line information may be inaccurate."
                )
            )

    # We should offer the Python error model for easier debugging of numerical
    # errors - e.g. division by zero.
    flags.error_model = "python"

    # Run compilation pipeline
    from numba_cuda_mlir.numba_cuda.core.target_extension import target_override
    from numba_cuda_mlir.lowering_utilities import context as mlir_ctx
    from numba_cuda_mlir._mlir import ir as mlir_ir

    flags.lto = is_lto

    with target_override("numba_cuda_mlir"), mlir_ctx.get_context(), mlir_ir.Location.unknown():
        cres = compiler.compile_extra(
            typingctx=typingctx,
            targetctx=targetctx,
            func=pyfunc,
            args=args,
            return_type=return_type,
            flags=flags,
            locals={},
            pipeline_class=get_compiler_class(targetoptions, launch_config_tracker),
        )

    cres.metadata.pop(_LAUNCH_CONFIG_TRACKER_METADATA_KEY, None)
    cres.metadata["targetoptions"] = targetoptions
    cres.metadata["gpu_target"] = gpu_target
    cres.metadata["transformed_source"] = transformed_source

    return cres


@global_compiler_lock
def mlir_compiler_entry(
    pyfunc,
    func_args,
    targetoptions: Dict[str, Any],
    abi="c",
    abi_info=None,
    override_argtypes=None,
):
    """Compile a Python function to MLIR for a given set of argument types.

    :param pyfunc: The Python function to compile.
    :param func_args: The function arguments used to invoke pyfunc.
    :param debug: Whether to include debug info in the compiled code.
    :type debug: bool
    :param lineinfo: Whether to include a line mapping from the compiled code
                     to the source code. Usually this is used with optimized
                     code (since debug mode would automatically include this),
                     so we want debug info in the LLVM IR but only the line
                     mapping in the final output.
    :type lineinfo: bool
    :param abi: The ABI for a compiled function - either ``"numba"`` or
                ``"c"``. Note that the Numba ABI is not considered stable.
                The C ABI is only supported for device functions at present.
    :type abi: str
    :param abi_info: A dict of ABI-specific options. The ``"c"`` ABI supports
                     one option, ``"abi_name"``, for providing the wrapper
                     function's name. The ``"numba"`` ABI has no options.
    :type abi_info: dict
    :param override_argtypes: Optional tuple of Numba types to use instead of
                              inferring from func_args. Used for tuple argument
                              flattening where the actual args don't match the
                              original tuple types.
    :type override_argtypes: tuple or None
    :return: (code, resty): The compiled code and inferred return type
    :rtype: tuple
    """
    if abi not in ("numba", "c"):
        raise NotImplementedError(f"Unsupported ABI: {abi}")

    abi_info = abi_info or dict()
    default_values = list(pyfunc.__defaults__) if pyfunc.__defaults__ else []
    sig = inspect.signature(pyfunc)

    # construct a list of remaining arguments of default values
    func_args_remained = default_values[
        len(func_args) + len(default_values) - len(sig.parameters) :
    ]
    func_args_extended = func_args + func_args_remained

    # Check if function has type annotations - if so, use declared signature
    # instead of inferring from runtime arguments
    params = list(sig.parameters.values())
    argtypes = []

    for i, param in enumerate(params):
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            if override_argtypes is not None:
                vararg_types = list(override_argtypes[i:])
            else:
                vararg_types = [typeof(a, Purpose.argument) for a in func_args_extended[i:]]
            argtypes.append(types.StarArgTuple.from_types(vararg_types))
            break

        override_argtype = (
            override_argtypes[i]
            if override_argtypes is not None and i < len(override_argtypes)
            else None
        )
        if isinstance(override_argtype, types.Literal):
            argtypes.append(override_argtype)
            continue

        annotation = param.annotation

        # If there's a Numba type annotation, use it directly
        if annotation != inspect.Parameter.empty and isinstance(annotation, types.Type):
            argtypes.append(annotation)
        else:
            # Try to convert annotation to Numba type, otherwise infer from runtime
            if annotation != inspect.Parameter.empty:
                try:
                    argtypes.append(to_numba_type(annotation))
                except (TypeError, NotImplementedError):
                    # Use override type if provided, otherwise infer from arg
                    if override_argtype is not None:
                        argtypes.append(override_argtype)
                    else:
                        argtypes.append(typeof(func_args_extended[i], Purpose.argument))
            else:
                # Use override type if provided, otherwise infer from arg
                if override_argtype is not None:
                    argtypes.append(override_argtype)
                else:
                    argtypes.append(typeof(func_args_extended[i], Purpose.argument))

    args, return_type = sigutils.normalize_signature(tuple(argtypes))

    cres = compile_mlir(
        pyfunc=pyfunc,
        return_type=return_type,
        args=args,
        targetoptions=targetoptions,
    )

    opt.optimize(cres)

    return cres



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception


from mlir.ir import DenseElementsAttr, Type

__all__ = [
  "QuantizedType",
  "AnyQuantizedType",
  "UniformQuantizedType",
  "UniformQuantizedPerAxisType",
  "CalibratedQuantizedType",
]

class QuantizedType(Type):
  @staticmethod
  def isinstance(type: Type) -> bool: ...

  @staticmethod
  def default_minimum_for_integer(is_signed: bool, integral_width: int) -> int:
    ...

  @staticmethod
  def default_maximum_for_integer(is_signed: bool, integral_width: int) -> int:
    ...

  @property
  def expressed_type(self) -> Type: ...

  @property
  def flags(self) -> int: ...

  @property
  def is_signed(self) -> bool: ...

  @property
  def storage_type(self) -> Type: ...

  @property
  def storage_type_min(self) -> int: ...

  @property
  def storage_type_max(self) -> int: ...

  @property
  def storage_type_integral_width(self) -> int: ...

  def is_compatible_expressed_type(self, candidate: Type) -> bool: ...

  @property
  def quantized_element_type(self) -> Type: ...

  def cast_from_storage_type(self, candidate: Type) -> Type: ...

  @staticmethod
  def cast_to_storage_type(type: Type) -> Type: ...

  def cast_from_expressed_type(self, candidate: Type) -> Type: ...

  @staticmethod
  def cast_to_expressed_type(type: Type) -> Type: ...

  def cast_expressed_to_storage_type(self, candidate: Type) -> Type: ...


class AnyQuantizedType(QuantizedType):

  @classmethod
  def get(cls, flags: int, storage_type: Type, expressed_type: Type,
          storage_type_min: int, storage_type_max: int) -> Type:
    ...


class UniformQuantizedType(QuantizedType):

  @classmethod
  def get(cls, flags: int, storage_type: Type, expressed_type: Type,
          scale: float, zero_point: int, storage_type_min: int,
          storage_type_max: int) -> Type: ...

  @property
  def scale(self) -> float: ...

  @property
  def zero_point(self) -> int: ...

  @property
  def is_fixed_point(self) -> bool: ...


class UniformQuantizedPerAxisType(QuantizedType):

  @classmethod
  def get(cls, flags: int, storage_type: Type, expressed_type: Type,
          scales: list[float], zero_points: list[int], quantized_dimension: int,
          storage_type_min: int, storage_type_max: int):
    ...

  @property
  def scales(self) -> list[float]: ...

  @property
  def zero_points(self) -> list[int]: ...

  @property
  def quantized_dimension(self) -> int: ...

  @property
  def is_fixed_point(self) -> bool: ...

class UniformQuantizedSubChannelType(QuantizedType):

  @classmethod
  def get(cls, flags: int, storage_type: Type, expressed_type: Type,
          scales: DenseElementsAttr, zero_points: DenseElementsAttr,
          quantized_dimensions: list[int], block_sizes: list[int],
          storage_type_min: int, storage_type_max: int):
    ...

  @property
  def quantized_dimensions(self) -> list[int]: ...

  @property
  def block_sizes(self) -> list[int]: ...

  @property
  def scales(self) -> DenseElementsAttr: ...

  @property
  def zero_points(self) -> DenseElementsAttr: ...

def CalibratedQuantizedType(QuantizedType):

  @classmethod
  def get(cls, expressed_type: Type, min: float, max: float): ...

  @property
  def min(self) -> float: ...

  @property
  def max(self) -> float: ...